Package: textmineR
Type: Package
Title: Functions for Text Mining and Topic Modeling
Version: 3.0.5
Authors@R: c(
    person("Tommy", "Jones", role = c("aut", "cre"), email = "jones.thos.w@gmail.com"),
    person("William", "Doane", email = "wil@drdoane.com", role = "ctb"),
    person("Mattias", "Attbom", email = "mattias1126@protonmail.com", role = "ctb")
    )
Description: An aid for text mining in R, with a syntax that
    should be familiar to experienced R users. Provides a wrapper for several 
    topic models that take similarly-formatted input and give similarly-formatted
    output. Has additional functionality for analyzing and diagnostics for
    topic models.
SystemRequirements: GNU make, C++11
Depends: R (>= 3.0.2), Matrix
Imports: gtools, magrittr, methods, parallel, text2vec (>= 0.5),
        stopwords, stringr, Rcpp, RcppProgress, RSpectra, utils
Suggests: spelling, digest, dplyr, igraph, knitr, lda, MASS, rmarkdown,
        SnowballC, stringi, testthat, tibble, tidyr, tidytext,
        topicmodels, wordcloud
License: MIT + file LICENSE
URL: https://www.rtextminer.com/
BugReports: https://github.com/TommyJones/textmineR/issues
LazyData: true
LinkingTo: Rcpp, RcppArmadillo, RcppProgress
RoxygenNote: 7.1.1
VignetteBuilder: knitr
Language: en-US
NeedsCompilation: yes
Packaged: 2021-06-27 23:50:12 UTC; tommy
Author: Tommy Jones [aut, cre],
  William Doane [ctb],
  Mattias Attbom [ctb]
Maintainer: Tommy Jones <jones.thos.w@gmail.com>
Repository: CRAN
Date/Publication: 2021-06-28 05:00:02 UTC
Built: R 4.0.5; x86_64-w64-mingw32; 2021-10-09 16:51:09 UTC; windows
Archs: x64
